/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.inovat.kex.bast.common.objectwrapper;

/*
 * Revision 1.2  2007/08/22 07:48:43  Trier
 * <Kein Kommentar eingegeben>
 */
import java.util.HashSet;
import java.util.List;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.inovat.kex.bast.common.DataModellMismatchException;
import de.inovat.kex.bast.common.NetzClientDavConnection;

/**
 * bietet Zugriff auf die {@link LangZeitZaehlStelle LangZeitZaehlStellen}, die
 * auch Achslasten erfassen.
 * Durch die Methode {@link #getAll} werden die DatenModelObjekte vom typ.achsLastMessStelle
 * gelesen und die {@link DEKanal.Axl} der jeweiligen {@link Fahrstreifen}
 * gesetzt.
 * 
 * @author LOGOS GmbH Rostock (www.logos-ib.de)
 * @author Ralf Trier
 * @version $Id: AxlLastMessStelle.java 442 2008-11-10 06:59:32Z Schawaller $
 */
public class AxlLastMessStelle {

  private static LangZeitZaehlStelle[] __LZZArray;


  public static boolean isNotInitialized() {
    return (__LZZArray == null);
  }

  /**
   * gibt die DatenModellObjekte vom typ langZeitZ&auml;hlstelle zurück, die
   * auch Achslasten erfassen.
   * DataModellMismatchException wird geworfen, wenn Objektreferenzen auf
   * falsche Typen verweisen bzw. diese nicht gesetzt sind.
   *
   * @throws DataModellMismatchException Exception.
   * @return LangZeitZaehlStelle[]
   */
  public static synchronized LangZeitZaehlStelle[] getAll() throws DataModellMismatchException {

    HashSet<LangZeitZaehlStelle> lzzSet = new HashSet<LangZeitZaehlStelle>();

    if (__LZZArray == null) {
      LangZeitZaehlStelle.getAll();

      DataModel model = NetzClientDavConnection.getInstance().getDataModel();

      SystemObjectType typeDeAxl = model.getType("typ.deAxl");
      SystemObjectType typeFahrstreifen = model.getType("typ.fahrStreifen");

      AttributeGroup atgAxlLastMessStelle = model.getAttributeGroup("atg.achsLastMessStelle");
      SystemObjectType soType = model.getType("typ.achsLastMessStelle");
      List<SystemObject> list = null;
      if (soType!=null) {
        list = soType.getElements();
      }

      if ((list!=null) && (list.size()>0)) {
        Data[] dataArray = model.getConfigurationData(list, atgAxlLastMessStelle);

        for (int i = 0; i < list.size(); i++) {

          ConfigurationObject confObject = (ConfigurationObject) list.get(i);
          Data oData = dataArray[i];

          SystemObject soAxlQuelle = oData.getReferenceValue("AchsLastMessStellenQuelle").getSystemObject();
          // System.err.println(soAxlQuelle.getType() + "soAxlQuelle=" + soAxlQuelle);
          DEKanal.Axl deKanal = null;
          if ((soAxlQuelle != null) && (soAxlQuelle.isOfType(typeDeAxl))) {
            deKanal = (DEKanal.Axl) DEKanal.getInstance((ConfigurationObject) soAxlQuelle);
          }
          else {
            throw new DataModellMismatchException("Quelle der achsLastMessStelle " + confObject.getPid() +
                                                  " ist nicht gesetzt oder nicht vom typ deAxl");
          }

          SystemObject soFS = oData.getReferenceValue("FahrStreifen").getSystemObject();
          // System.err.println(soFS.getType() + "soFS=" + soFS);
          Fahrstreifen fs = null;
          if ((soFS != null) && (soFS.isOfType(typeFahrstreifen))) {
            fs = Fahrstreifen.getInstance((ConfigurationObject) soFS);
          }
          else {
            throw new DataModellMismatchException("Fahrstreifen der achsLastMessStelle " + confObject.getPid() +
                                                  " nicht gesetzt oder nicht vom typ fahrStreifen");
          }
          fs.setDEKanalAxl(deKanal);
          LangZeitZaehlStelle lzz = fs.getMessQuerschnitt().getAuswerteQuerschnitt().getLangZeitZaehlStelle();

          lzzSet.add(lzz);
        }
      }
      __LZZArray = lzzSet.toArray(new LangZeitZaehlStelle[lzzSet.size()]);
    }
    return __LZZArray;
  }


}
