/*
 * Copyright (c) 2010 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.kex.bast.importer.BAStImportSteuerung
 *
 * de.inovat.kex.bast.importer.BAStImportSteuerung is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.kex.bast.importer.BAStImportSteuerung is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.kex.bast.importer.BAStImportSteuerung; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Köln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */

package de.inovat.kex.bast.importer;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.archive.*;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.inovat.kex.bast.KonstantenBast.BAST_FORMAT;
import de.inovat.kex.bast.SendeBAStStatusDaten;
import de.inovat.kex.bast.common.NetzClientDavConnection;
import de.inovat.kex.bast.common.objectwrapper.Fahrstreifen;
import de.inovat.kex.bast.common.objectwrapper.LangZeitZaehlStelle;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

//~ JDK IMPORTE ===============================================================

//~ KLASSEN ===================================================================

/**
 * Die Klasse dient der Importsteuerung der BASt-Band Daten.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dagmar Rostek (DR)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BAStImportSteuerung implements StandardApplication, ClientReceiverInterface, ClientSenderInterface, Runnable {

    /** Geschwindigkeitsgrenzen BPA. */
    private static final int[] AR_bPAValues = { 0, 40, 50, 60, 70, 80, 90, 100, 110, 120 };

    /** Array mit DaV-Attributen für atg.verkehrsDatenLangZeitIntervall. */
    private static final String[] AR_verkehrsDatenLangZeitIntervall = { "qKfz", "qPkwÄ", "qKfzNk", "qPkwG", "qPkw", "qKrad", "qLfw", "qLkwÄ", "qPkwA", "qLkw", "qBus", "qLkwK", "qLkwA", "qSattelKfz", "vKfz", "vPkwÄ", "vKfzNk", "vPkwG", "vPkw", "vKrad", "vLfw", "vLkwÄ", "vPkwA", "vLkw", "vBus", "vLkwK", "vLkwA", "vSattelKfz", "sKfz", "sPkwÄ", "sKfzNk", "sPkwG", "sPkw", "sKrad", "sLfw", "sLkwÄ", "sPkwA", "sLkw", "sBus", "sLkwK", "sLkwA", "sSattelKfz", "v85Kfz", "v85PkwÄ", "v85KfzNk", "v85PkwG",
                                                                        "v85Pkw", "v85Krad", "v85Lfw", "v85LkwÄ", "v85PkwA", "v85Lkw", "v85Bus", "v85LkwK", "v85LkwA", "v85SattelKfz" };

    /** Geschwindigkeitsgrenzen LVO. */
    private static final int[] AR_lVoValues = { 0, 40, 50, 60, 70, 80, 90, 100, 110, 120, 130, 140, 150, 160, 170, 180 };

    /** Geschwindigkeitsgrenzen SGV. */
    private static final int[] AR_sGVValues = { 0, 40, 50, 60, 70, 80, 90, 100, 110, 120 };

    /**
     * Array mit allen Fahrzeugarten nach TLS 2002 die nicht über den Import versorgt werden, aber mit Daten für das
     * Versenden im Datenverteiler versorgt werden müssen.
     */
    private static final String[] AR_VerkehrsDatenLangZeitIntervall_WertGeschwKlassen = { "qKfzGeschwKlasse", "qPkwÄGeschwKlasse", "qKfzNkGeschwKlasse", "qPkwGGeschwKlasse", "qPkwGeschwKlasse", "qKradGeschwKlasse", "qLfwGeschwKlasse", "qLkwÄGeschwKlasse", "qPkwAGeschwKlasse", "qLkwGeschwKlasse", "qBusGeschwKlasse", "qLkwKGeschwKlasse", "qLkwAGeschwKlasse", "qSattelKfzGeschwKlasse" };

    /**
     * ToDo HCK: geändert 25.11.11, anteilSGV muss raus, sonst wird Wert beim Import auf nicht ermittelbar gesetzt:
     * private String[] AR_VerkehrsDatenLangZeitIntervallGeschwVerteilungFS = { "vBPA", "sBPA", "v15BPA", "v85BPA", "qSGV", "anteilSGV", "vLVo", "sLVo", "v15LVo", "v85LVo", "vSGV", "sSGV", "v15SGV", "v85SGV" };
     * Array mit der Teilmenge der Geschiwndigkeitsdaten.
     */
    private static final String[] AR_VerkehrsDatenLangZeitIntervallGeschwVerteilungFS = { "vBPA", "sBPA", "v15BPA", "v85BPA", "qSGV", "vLVo", "sLVo", "v15LVo", "v85LVo", "vSGV", "sSGV", "v15SGV", "v85SGV" };

    /**
     * Array mit allen Attributen der Attributgruppe atg.verkehrsDatenLangZeitIntervallGeschwindigkeitVerteilungFS die
     * weder exportiert, noch importiert werden.
     */
    private static final String[] AR_VerkehrsDatenLangZeitIntervallGeschwVerteilungFSnichtImportierteAttribute = { "qKfzNk", "qKrad", "qPkw", "qLfw", "qPkwA", "qLkw", "qLkwA", "qSattelKfz", "qBus", "qSGV", "anteilSGV" };

    /** Klassen. */
    private static final String[] AR_vLVoSGvBPa = { "vLVo", "sLVo", "v15LVo", "v85LVo", "vSGV", "sSGV", "v15SGV", "v85SGV", "vBPA", "sBPA", "v15BPA", "v85BPA" };

    /** Aspekt Anfrage. */
    private static final String ASP_ANFRAGE = "asp.anfrage";

    /** Aspekt ExterneErfassung für die Archivanfrage der V2004 Daten. */
    private static final String ASP_EXTERNE_ERFASSUNG = "asp.externeErfassung";

    /** Aspekt MessWertErsetzung. */
    private static final String ASP_MESSWERT_ERSETZUNG = "asp.messWertErsetzung";

    /** Aspekt des Systemobjektes Importverzeichnisüberwachung. */
    private static final String ASP_ZUSTAND = "asp.zustand";

    /** Attributgruppe zur Steuerung des BASt-Band Imports: */
    private static final String ATG_IMPORT_STEUERUNG = "atg.bastImportSteuerung";

    /** Attributgruppe Ergebnismeldung Version 24 (BASt 2007). */
    private static final String ATG_VERKEHRS_DAT_LANG_ZEIT_INT_GESCHW_VERTEILUNG = "atg.verkehrsDatenLangZeitIntervallGeschwindigkeitVerteilungFs";

    /** Attributgruppe Verkehrsdatenlangzeitinervall (BASt 2004). */
    private static final String ATG_VERKEHRS_DATEN_LANG_ZEITINTERVALL = "atg.verkehrsDatenLangZeitIntervall";

    /** Attributgruppe des Systemobjektes Importverzeichnisüberwachung. */
    private static final String   ATG_VERZEICHNIS_UEBERWACHUNG = "atg.bastImportVerzeichnisUeberwachung";
    protected static final String DATEI_BEARBEITET             = ".bearbeitet";
    protected static final String DATEI_LOCK                   = ".lock";
    private static final String   ERFASSUNG                    = "Erfassung";

    /** TODO Fileseparator auf dem Server. */
    protected final static String FILESEPARATOR_SERVER = "/";
    private static final String   GUETE                = "Güte";
    private static final String   IMPLAUSIBEL          = "Implausibel";
    private static final String   INDEX                = "Index";
    private static final String   INTERPOLIERT         = "Interpoliert";
    private static final String   JA                   = "Ja";
    private static final String   MESSWERT_ERSETZUNG   = "MessWertErsetzung";
    private static final String   NICHT_ERFASST        = "NichtErfasst";
    private static final String   NICHT_ERMITTELBAR    = "nicht ermittelbar";
    private static final String   PL_Formal            = "PlFormal";
    private static final String   Q_BPAGeschwKlasse    = "qBPAGeschwKlasse";
    private static final String   Q_LVoGeschwKlasse    = "qLVoGeschwKlasse";
    private static final String   Q_SGVGeschwKlasse    = "qSGVGeschwKlasse";
    private static final String   STANDARD             = "Standard";
    private static final String   STATUS               = "Status";

    /** Debugdebug fuer Debug-Ausgaben. */
    private static Debug        debug            = Debug.getLogger();
    private static final String VERFAHREN        = "Verfahren";
    private static final String WERT             = "Wert";
    private static final String WERT_MAX         = "WertMax";
    private static final String WERT_MAX_LOGISCH = "WertMaxLogisch";
    private static final String WERT_MIN         = "WertMin";
    private static final String WERT_MIN_LOGISCH = "WertMinLogisch";

    //~ FELDER ================================================================

    /** Integer Attribut dass den Prozentsatz des aktuellen Importes repräsentiert. */
    private int _prozentImport = 2;

    /** Boolesches Attribut, das angibt, ob der Thread gerade aktiv ist. */
    private boolean _threadImportAktiv = false;

    /** HashMap mit den Sendestatus für das DataDescription des SystemObjekts. */
    private Map<String, Boolean> _mapSendeErlaubnis = new HashMap<String, Boolean>();

    /**
     * Boolesches Attribut das angibt, ob der Importvorgang durch das System abgebrochen werden musste. Dies ist z.B.
     * der Fall, wenn kein Archivdatensatz zu dem zu importierendem Datensatz vorliegt.
     */
    private boolean _abbruchSystem = false;

    /** Boolesches Attribut das angibt, ob der Import unterbrochen wurde. */
    private boolean _abbruchImport = false;

    /** Variable für die TESTZWECKE (true). Im Normalbetrieb wird auf false gesetzt. */
    private boolean _istFuerTestzwecke = false;

    /** Attributgruppe VerkehrsDatenLangZeitIntervall wird für den Import der BASt-Version 2004 benötigt. */
    private AttributeGroup _atgVerkehrsDatenLangZeitIntervall;

    /** Attributgruppe tlsLveErgebnismeldungVersion24, wird für den Import BASt-Version 2007 benötigt. */
    private AttributeGroup _atgVerkehrsDatLangZeitIntGeschwVerteilungFS;

    /** Importformat. */
    private BAST_FORMAT _bastFormat;

    /** Statusinformationen die an den Datenverteiler gesendet werden. */
    private SendeBAStStatusDaten _bastStatusDaten;

    /** Datenbeschreibung für das Versenden der Importdaten. */
    private DataDescription _datenbeschreibungVerkehrsDatenLangZeitIntervall;

    /** Datenbeschreibung für V2007. */
    private DataDescription _datenbeschreibungVerkehrsDatLangZeitIntGeschwVerteilungFS;

    /** ClientDavInteface. */
    private ClientDavInterface _dav;

    /** FTP-Benutzername. */
    private String _ftpBenutzer;

    /** Legt fest, ob die FTP-Verbindung im PassivModus eingebaut werden soll. */
    private boolean _ftpPassivModus;

    /** FTP-Passwort. */
    private String _ftpPasswd;

    /** Host des FTP-Servers. */
    private String _ftpServerName;

    /** Port des FTP-Servers. */
    private int _ftpServerPort;

    /** ArrayList der Dateinamen die importiert werden sollen. */
    private ArrayList<String> _importDateienList;

    /**
     * Definiert den ServerTyp: FTP oder SFTP:
     * falls true - FTP,
     * sonst - SFTP.
     */
    private boolean _istFtp;

    /** Objekt vom Typ Langzeitzählstelle. */
    private LangZeitZaehlStelle _lzz;

    /** Klasse für die Kommunikation mit dem Server. */
    private IServerKommunikation _serverKommunikation;

    /** Lokaler Temp-Ordner. */
    private String _tempOrdner;

    /** Thread für die BastBandImport Steuerung. */
    private Thread _threadBastBandImport;

    /** Typ des Systemobjekte für den BASt-Band Import/Export. */
    private String _typBAStBand;

    //~ METHODEN ==============================================================

    /**
     * Die Methode stoppt den Thread.
     *
     * @param meldung Meldung die dem Clienten mitgeteilt wird.
     */
    private void abbrechen(String meldung) {
        _abbruchImport     = true;
        _threadImportAktiv = false;
        sendeMeldungInfo("100", meldung);
    }

    /**
     * Anmeldung beim Datenverteiler als Empfänger für die Attribute
     * bzgl. BASt-Band Import und Verzeichnisüberwachung.
     */
    protected void anmeldenDavEmpfaengerBAStImport() {

        // ---------------------------------------------------------------
        // Initialisiere DaV-Daten.
        // ---------------------------------------------------------------
        _atgVerkehrsDatenLangZeitIntervall = _dav.getDataModel().getAttributeGroup(ATG_VERKEHRS_DATEN_LANG_ZEITINTERVALL);

        Aspect aspMessWertErsetzung = _dav.getDataModel().getAspect(ASP_MESSWERT_ERSETZUNG);

        _datenbeschreibungVerkehrsDatenLangZeitIntervall = new DataDescription(_atgVerkehrsDatenLangZeitIntervall, aspMessWertErsetzung);

        // ---------------------------------------------------------------
        _atgVerkehrsDatLangZeitIntGeschwVerteilungFS               = _dav.getDataModel().getAttributeGroup(ATG_VERKEHRS_DAT_LANG_ZEIT_INT_GESCHW_VERTEILUNG);
        _datenbeschreibungVerkehrsDatLangZeitIntGeschwVerteilungFS = new DataDescription(_atgVerkehrsDatLangZeitIntGeschwVerteilungFS, aspMessWertErsetzung);

        // ---------------------------------------------------------------
        // DaV-ANMELDUNG
        // ---------------------------------------------------------------
        // Initialisiere SystemObjekt Systemobjekt für den BASt-Band Import:
        SystemObject soBAStImportSteuerung = _dav.getDataModel().getObject(_typBAStBand);

        if (soBAStImportSteuerung == null) {
            throw new RuntimeException(String.format("Das Systemobjekt existiert nicht <%s>.", _typBAStBand));
        }

        // -------------------------------------------------------------
        // BASt-Band Import:
        // Anmelden zum Empfang der Daten
        AttributeGroup  atg                              = _dav.getDataModel().getAttributeGroup(ATG_IMPORT_STEUERUNG);
        Aspect          asp                              = _dav.getDataModel().getAspect(ASP_ANFRAGE);
        DataDescription datenbeschreibungImportSteuerung = new DataDescription(atg, asp);

        try {
            _dav.subscribeReceiver(this, soBAStImportSteuerung, datenbeschreibungImportSteuerung, ReceiveOptions.normal(), ReceiverRole.drain());
        }
        catch (Exception e) {
            throw new RuntimeException("Probleme bei der Anmeldung beim Datenverteiler. Die Anwendung wird geschlossen.", e);
        }

        // -------------------------------------------------------------
        // Verzeichnisüberwachung:
        // Anmelden zum Empfang der Daten
        atg = _dav.getDataModel().getAttributeGroup(ATG_VERZEICHNIS_UEBERWACHUNG);
        asp = _dav.getDataModel().getAspect(ASP_ZUSTAND);

        DataDescription datenbeschreibungVerzeichnisUeberwachung = new DataDescription(atg, asp);

        try {
            _dav.subscribeReceiver(this, soBAStImportSteuerung, datenbeschreibungVerzeichnisUeberwachung, ReceiveOptions.normal(), ReceiverRole.drain());
        }
        catch (Exception e) {
            throw new RuntimeException("Probleme bei der Anmeldung beim Datenverteiler. Die Anwendung wird geschlossen.", e);
        }
    }

    /**
     * Die Methode ermittelt zu einem Fahrstreifen und dem aktuellen Stundendatensatz des Fahrstreifens  die aktuellsten
     * Archivdaten, diese werden in einer Map gespeichert, in der der Schlüssel der Zeitpunkt, der Wert das Dataobjekt
     * ist. Dadurch, dass die Zeit als Schlüssel verwendet wird, wird gewährleistet, dass immer der aktuellste Datensatz
     * im Archiv verwendet wird.
     *
     * @param arListStundenDatenSatzfuerFS Liste aller Stundendatensätze zu dem Fahrstreifenobjekt.
     * @param fs                           Fahrstreifenobjekt.
     * @param atg                          Attributgruppe für die der Archivdatensatz bestimmt wird.
     *
     * @return Map mit den Dataobjekten zu einem Stundendatensatz des Fahrstreifens.
     *
     * @throws IOException          .
     * @throws InterruptedException .
     */
    private HashMap<Long, Data> bestimmeArchivStundendatensatzMap(ArrayList<Stundendatensatz> arListStundenDatenSatzfuerFS, Fahrstreifen fs, AttributeGroup atg) throws IOException, InterruptedException {

        // Aspekt für die Archivanfrage.
        Aspect aspExterneErfassung = _dav.getDataModel().getAspect(ASP_EXTERNE_ERFASSUNG);

        // HashMap mit den letzten Stundendatensätzen.
        HashMap<Long, Data> archivDataZeitObj = new HashMap<Long, Data>();

        /** Array Archivdatastream. */
        ArchiveDataStream[] archiveDataStream;

        // Archivanfrageergebnis.
        ArchiveDataQueryResult queryResult;
        ArchiveRequestManager  archiveRequestManager = _dav.getArchive();

        // Es werden nur die Daten vom Typ ArchiveDataKind.ONLINE angefordert.
        ArchiveDataKindCombination archiveDataKindCombination = new ArchiveDataKindCombination(ArchiveDataKind.ONLINE);

        // Liste der Archivspezifikation für die Anfrage.
        List<ArchiveDataSpecification> archiveDataSpecifications = new LinkedList<ArchiveDataSpecification>();

        // Datenbeschreibung der Archivanfrage.
        DataDescription dataDescription = new DataDescription(atg, aspExterneErfassung);

        // Startzeitpunkt für die Archivanfrage, d.h. es wird die erste Datenzeit des Stundendatensatzes als Startzeit verwendet.
        long startZeit = 0;

        // Endzeitpunkt für die Archivanfrage, d.h. es wird die Datenzeit des letzten Stundendatensatzes als Endzeit verwendet.
        long endZeit = 0;

        // Bestimmung der Start- und Endzeit für die Archivanfrage, dabei wird der akuelle Stundendatensatz betrachtet und entsprechend die Archivanfrage erstellt.
        if ((arListStundenDatenSatzfuerFS != null)
            && (arListStundenDatenSatzfuerFS.size() > 0)) {
            int laengeStundendatensaetzeFS = arListStundenDatenSatzfuerFS.size();

            startZeit = arListStundenDatenSatzfuerFS.get(0).getDatumZeit();
            endZeit   = arListStundenDatenSatzfuerFS.get(laengeStundendatensaetzeFS - 1).getDatumZeit();
        }

        // Erzeugung ArchiveTimeSpecification mit aktuellen Start und Endzeitpunkten.
        ArchiveTimeSpecification archiveTimeSpecification = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, startZeit, endZeit);

        archiveDataSpecifications.add(new ArchiveDataSpecification(archiveTimeSpecification, archiveDataKindCombination, ArchiveOrder.BY_DATA_TIME, ArchiveRequestOption.NORMAL, dataDescription, fs.getConfigurationObject()));

        // Archivanfrageergebnis.
        queryResult = archiveRequestManager.request(ArchiveQueryPriority.MEDIUM, archiveDataSpecifications);

        // Auswertung der Importierten Daten zuzüglich der Archivanfrage Daten.
        if (archiveRequestManager.isArchiveAvailable()) {
            if (queryResult.isRequestSuccessful()) {
                archiveDataStream = queryResult.getStreams();

                for (ArchiveDataStream anArchiveDataStream : archiveDataStream) {
                    ArchiveData archivDaten = anArchiveDataStream.take();

                    while (archivDaten != null) {
                        Data archivDataObj = archivDaten.getData();

                        if ((archivDataObj != null)) {
                            archivDataZeitObj.put(archivDaten.getDataTime(), archivDataObj);
                        }

                        archivDaten = anArchiveDataStream.take();
                    }
                }
            }
        } else {
            _abbruchSystem = true;
            sendeMeldungFehler("100", "Das Archiv ist nicht verfügbar.");
        }

        return archivDataZeitObj;
    }

    /**
     * Um in der Bedienung und Visualisierung den aktuellen Status über den Importfortschritt anzuzeigen wird ein Wert
     * übermittelt, der je nach Prozentschritt - in Abhängigkeit der zu importierenden Daten - die Prozentzahl der
     * Progressbar um eins erhöht.
     *
     * @param countStundendatensaetze Anzahl der aktuell bearbeiteten Stundendatensätze.
     * @param prozentSchritt          Prozentschritt.
     *
     * @return Anzahl der aktuell bearbeitetend Stundendatensätze, sobald die Anzahl der abgearbeiteten
     *         Stundendatensätze gleich dem prozentSchritt ist, wird die Prozentzahl für den Import um eins erhöht und
     *         der counter für die Stundendatensätze auf 0 initialisiert.
     */
    private int bestimmeProgressbarFortschritt(int countStundendatensaetze, int prozentSchritt) {
        if (countStundendatensaetze == prozentSchritt) {
            try {
                if (_prozentImport < 100) {
                    _prozentImport += 1;
                }

                _bastStatusDaten.sendeBAStStatusDaten((new Integer(_prozentImport)).toString(), "");
            }
            catch (SendSubscriptionNotConfirmed e) {
                debug.error("Es sind Probleme beim Senden der Statusinformationsdaten aufgetreten.", e);
            }

            countStundendatensaetze = 0;
        }

        return countStundendatensaetze;
    }

    /**
     * Die Methode ermittelt den aktuellen Prozentstand.
     *
     * @return Prozentschritt je nach Importdateigrösse.
     */
    private int bestimmeProzentSchritt() {
        int anzahlFSInRichtung    = _lzz.getInAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen().length;
        int anzahlFSGegenRichtung = _lzz.getGegenAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen().length;
        int anzahlFS              = anzahlFSInRichtung + anzahlFSGegenRichtung;
        int prozentSchritt;

        if (_importDateienList.size() == 1) {
            double prozent = (double) anzahlFS * 7.5;

            prozentSchritt = (int) prozent;
        } else {
            double prozent = (double) anzahlFS * 7.2 * (double) _importDateienList.size();

            prozentSchritt = (int) prozent;
        }

        return prozentSchritt;
    }

    @Override
    public void dataRequest(SystemObject so, DataDescription datenBeschreibung, byte status) {
        _mapSendeErlaubnis.put(generiereSendeErlaubnisSchluessel(so, datenBeschreibung), (status == START_SENDING));
    }

    /**
     * Die Methode erzeugt einen Datensatz und übergibt diesen zum Senden an den Dav. Dabei wird die Attributgruppe
     * atg.VerkehrsdatenLangZeitIntervall mit dem Aspekt asp.messWertErsetzung geschrieben. Um einen vollständigen
     * Datensatz zu erzeugen, wird der zugehörige Archivdatensatz abgerufen asp.ExterneErfassung und als Default
     * verwendet, d.h. nur die zusätzlich importierten Daten werden in dem Archivdatensatz entsprechend angepasst, dann
     * aber via Aspekt asp.messWertErsetzung publiziert.
     *
     * Methode ist public für die TESTZWECKE.
     *
     * @param bastFormat2004 BastBand Format.
     * @param fs             Fahrstreifen.
     * @param prozentSchritt Prozentschritt.
     *
     * @throws Exception .
     */
    public void erzeugeDatensatzV2004(BastBestandsbandformat bastFormat2004, Fahrstreifen fs, int prozentSchritt) throws Exception {

        // Liste ResultData zum Versenden.
        ArrayList<ResultData> listeResultData = new ArrayList<ResultData>();

        // Zähler zum setzten der Progressbar entsprechend des Importfortschrittes.
        int countStundendatensaetze = 0;

        // Liste aller Stundendatensaetze.
        ArrayList<Stundendatensatz> arListStundenDatenSatzfuerFS = bastFormat2004.getMapFSArrayListStundendatenSaetze8Plus1().get(fs);

        // HashMap mit den aktuellsten Archivdatensätzen zu einem Stundendatensatz.
        HashMap<Long, Data> archivDataZeitObj;

        if (!_istFuerTestzwecke) {
            archivDataZeitObj = bestimmeArchivStundendatensatzMap(arListStundenDatenSatzfuerFS, fs, _atgVerkehrsDatenLangZeitIntervall);
        } else {
            archivDataZeitObj = new HashMap<Long, Data>();
        }

        for (Stundendatensatz stundenSatz : arListStundenDatenSatzfuerFS) {
            HashMap<Long, Data> aktuellstesDataObj             = new HashMap<Long, Data>();
            Data                dataObjVorbelegtMitImportDaten = erzeugeVerkehrsDatenLangZeitIntervallDatensatzMitVorbelegungImportdaten(stundenSatz);
            Map<String, String> mapFahrzeugArtenWert           = stundenSatz.getMapFZArtVerkehrsMenge();
            Data                archivStundenDatenSatz         = archivDataZeitObj.get(stundenSatz.getDatumZeit());

            if (archivStundenDatenSatz == null) {
                Calendar cal = Calendar.getInstance();

                cal.setTimeInMillis(stundenSatz.getDatumZeit());

                // D.h. kein Stundendatensatz für dieses Fahrstreifenobjekt gefunden. Infomeldung an BuV
                if (!_istFuerTestzwecke) {
                    sendeMeldungFehler("", String.format("Für das Fahrstreifenobjekt %s existiert kein archivierter Stundendatensatz für den Zeitpunkt %s. Die Werte werden mit nicht ermittelbar vorbelegt!", fs.getPID(), cal.getTime().toString()));
                }

                aktuellstesDataObj.put(stundenSatz.getDatumZeit(), dataObjVorbelegtMitImportDaten);

                if (_istFuerTestzwecke) {
                    if (dataObjVorbelegtMitImportDaten.toString().contains("Undefiniert")) {
                        debug.error(String.format("Mindenstenz ein Datensatz enthält undefinierte Daten: %s %s %s", stundenSatz.getDatumZeit(), fs, dataObjVorbelegtMitImportDaten));
                        System.exit(-1);
                    }
                }
            } else {
                for (String fzArt : mapFahrzeugArtenWert.keySet()) {
                    Data dataWert = archivStundenDatenSatz.getItem(fzArt);

                    // Aus einer map zu einem Fahrstreifen den Wert auslesen.
                    if (!mapFahrzeugArtenWert.get(fzArt).equals("")) {

                        // Füllen das Archivdatensatzes mit den aktuell importierten Daten
                        dataWert.getTextValue(WERT).setText(mapFahrzeugArtenWert.get(fzArt));
                    }
                }

                aktuellstesDataObj.put(stundenSatz.getDatumZeit(), archivStundenDatenSatz);
            }

            countStundendatensaetze++;
            countStundendatensaetze = bestimmeProgressbarFortschritt(countStundendatensaetze, prozentSchritt);

            if (!aktuellstesDataObj.isEmpty()) {

                // Telegramm erstellen
                listeResultData.add(new ResultData(_dav.getDataModel().getObject(fs.getConfigurationObject().getPid()), _datenbeschreibungVerkehrsDatenLangZeitIntervall, stundenSatz.getDatumZeit(), aktuellstesDataObj.get(stundenSatz.getDatumZeit())));
            }
        }

        if (!_istFuerTestzwecke) {
            sendeDavDaten(listeResultData);
        }
    }

    /**
     * Die Methode erzeugt einen BASt-Band Bestandsformat Daten V2007 Datensatz und leitet das Senden dieses Datensatzes
     * ein, wenn erfolgreich ein Datensatz erzeugt wurde.
     *
     * @param bastBestandsFormat BASt-Band Format V2007
     * @param fs                 Fahrstreifen.
     * @param prozentSchritt     Prozentschritt.
     *
     * @throws Exception .
     */
    private void erzeugeDatensatzV2007(BastBestandsbandformat bastBestandsFormat, Fahrstreifen fs, int prozentSchritt) throws Exception {

        // Liste ResultData zum Versenden.
        ArrayList<ResultData> listeResultData = new ArrayList<ResultData>();

        // Zähler zum setzten der Progressbar entsprechend des Importfortschrittes.
        int countStundendatensaetze = 0;

        // Liste aller Stundendatensaetze.
        ArrayList<Stundendatensatz> arListStundenDatenSatzfuerFS = bastBestandsFormat.getMapFSArrayListStundendatenSaetze8Plus1().get(fs);

        sendeMeldungInfo("", "Die Archivanfrage für das Fahrstreifenobjekt " + fs.getPID() + " wird erstellt.");

        HashMap<Long, Data> archivDataZeitObj = bestimmeArchivStundendatensatzMap(arListStundenDatenSatzfuerFS, fs, _atgVerkehrsDatLangZeitIntGeschwVerteilungFS);

        sendeMeldungInfo("", "Die Auswertung des Stundendatensatzes wird gestartet ...");

        for (Stundendatensatz stundenSatz : arListStundenDatenSatzfuerFS) {
            HashMap<Long, Data> aktuellstesDataObj             = new HashMap<Long, Data>();
            int                 vkLVo_size                     = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("LVo").size();
            int                 vkSGV_size                     = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("SGV").size();
            int                 vkBPA_size                     = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("BPA").size();
            Data                dataObjVorbelegtMitImportDaten = erzeugeVerkehrsDatenLangZeitIntervallGeschwindigkeitVerteilungDatensatzMitVorbelegungImportdaten(stundenSatz);
            Map<String, String> mapFahrzeugArtenWert           = stundenSatz.getMapFZArtVerkehrsMenge();
            Data                archivStundenDatenSatz         = archivDataZeitObj.get(stundenSatz.getDatumZeit());

            if (archivStundenDatenSatz == null) {
                Calendar cal = Calendar.getInstance();

                cal.setTimeInMillis(stundenSatz.getDatumZeit());

                // D.h. kein Stundendatensatz für dieses Fahrstreifenobjekt gefunden. Infomeldung an BuV
                sendeMeldungFehler("", String.format("Für das Fahrstreifenobjekt %s existiert kein archivierter Stundendatensatz für den Zeitpunkt %s. Die Werte werden mit nicht ermittelbar vorbelegt!", fs.getPID(), cal.getTime().toString()));
                aktuellstesDataObj.put(stundenSatz.getDatumZeit(), dataObjVorbelegtMitImportDaten);
            } else {
                for (String fzArt : AR_VerkehrsDatenLangZeitIntervallGeschwVerteilungFS) {
                    if (mapFahrzeugArtenWert.keySet().contains(fzArt)) {

                        // Aus einer map zu einem Fahrstreifen den Wert auslesen.
                        if (!mapFahrzeugArtenWert.get(fzArt).equals("")) {

                            // Füllen das Archivdatensatzes mit den aktuell importierten Daten
                            archivStundenDatenSatz.getTextValue(fzArt).setText(mapFahrzeugArtenWert.get(fzArt));
                        }
                    } else {
                        archivStundenDatenSatz.getTextValue(fzArt).setText(NICHT_ERMITTELBAR);
                    }
                }

                archivStundenDatenSatz.getArray(Q_LVoGeschwKlasse).setLength(vkLVo_size);

                for (int i = 0; i < vkLVo_size; i++) {
                    archivStundenDatenSatz.getUnscaledArray(Q_LVoGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("LVo").get(AR_lVoValues[i]));
                }

                archivStundenDatenSatz.getArray(Q_SGVGeschwKlasse).setLength(vkSGV_size);

                for (int i = 0; i < vkSGV_size; i++) {
                    archivStundenDatenSatz.getUnscaledArray(Q_SGVGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("SGV").get(AR_sGVValues[i]));
                }

                archivStundenDatenSatz.getArray(Q_BPAGeschwKlasse).setLength(vkBPA_size);

                for (int i = 0; i < vkBPA_size; i++) {
                    archivStundenDatenSatz.getUnscaledArray(Q_BPAGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("BPA").get(AR_bPAValues[i]));
                }

                aktuellstesDataObj.put(stundenSatz.getDatumZeit(), archivStundenDatenSatz);
            }

            countStundendatensaetze++;
            countStundendatensaetze = bestimmeProgressbarFortschritt(countStundendatensaetze, prozentSchritt);

            if (!aktuellstesDataObj.isEmpty()) {

                // Telegramm erstellen
                listeResultData.add(new ResultData(_dav.getDataModel().getObject(fs.getConfigurationObject().getPid()), _datenbeschreibungVerkehrsDatLangZeitIntGeschwVerteilungFS, stundenSatz.getDatumZeit(), aktuellstesDataObj.get(stundenSatz.getDatumZeit())));
            }
        }

        sendeDavDaten(listeResultData);
    }

    /**
     * Die Methode erzeugt zu einem Stundendatensatz ein mit "Nicht ermittelbar" vorbelegten Werten, wobei alle aus der
     * Datei importierten Werte entsprechend belegt werden.
     *
     * @param stundenSatz Stundendatensatz, für den der Datensatz erzeugt wird.
     *
     * @return data Objekt.
     */
    private Data erzeugeVerkehrsDatenLangZeitIntervallDatensatzMitVorbelegungImportdaten(Stundendatensatz stundenSatz) {
        Data dataVDatenLangZeitIntervall = _dav.createData(_atgVerkehrsDatenLangZeitIntervall);

        dataVDatenLangZeitIntervall.getTimeValue("T").setText("1 Stunde");

        Map<String, String> mapFahrzeugArtenWert = stundenSatz.getMapFZArtVerkehrsMenge();

        for (String att : AR_verkehrsDatenLangZeitIntervall) {
            Data dataWert = dataVDatenLangZeitIntervall.getItem(att);

            // Aus einer map zu einem Fahrstreifen den Wert auslesen
            boolean istVorhanden = (mapFahrzeugArtenWert.get(att) != null)
                                   &&!mapFahrzeugArtenWert.get(att).isEmpty();

            if (istVorhanden) {
                dataWert.getTextValue(WERT).setText(mapFahrzeugArtenWert.get(att));
            } else {
                dataWert.getTextValue(WERT).setText(NICHT_ERMITTELBAR);
            }

            Data dataStatus    = dataWert.getItem(STATUS);
            Data dataErfassung = dataStatus.getItem(ERFASSUNG);

            dataErfassung.getTextValue(NICHT_ERFASST).setText(JA);

            Data dataPIFormal = dataStatus.getItem(PL_Formal);

            dataPIFormal.getTextValue(WERT_MAX).setText(JA);
            dataPIFormal.getTextValue(WERT_MIN).setText(JA);

            Data dataPILogisch = dataStatus.getItem("PlLogisch");

            dataPILogisch.getTextValue(WERT_MAX_LOGISCH).setText(JA);
            dataPILogisch.getTextValue(WERT_MIN_LOGISCH).setText(JA);

            Data dataMessWertErsetzung = dataStatus.getItem(MESSWERT_ERSETZUNG);

            dataMessWertErsetzung.getTextValue(IMPLAUSIBEL).setText(JA);
            dataMessWertErsetzung.getTextValue(INTERPOLIERT).setText(JA);

            Data dataguete = dataWert.getItem(GUETE);

            //
            if (istVorhanden) {
                dataguete.getScaledValue(INDEX).setText(NICHT_ERMITTELBAR);
            } else {
                dataguete.getScaledValue(INDEX).set(1);
            }

            dataguete.getScaledValue(VERFAHREN).setText(STANDARD);
        }

        // Daten existieren nicht, also wird die Länge des Array's jeweils auf 0 gesetzt.
        for (String nameGKlassen : AR_VerkehrsDatenLangZeitIntervall_WertGeschwKlassen) {
            dataVDatenLangZeitIntervall.getArray(nameGKlassen).setLength(0);
        }

        return dataVDatenLangZeitIntervall;
    }

    /**
     * Die Methode erzeugt einen Stundendatensatz der mit den Importdaten vorbelegt ist. Alle Werte die der
     * Importdatensatz nicht liefert werden auf "nicht ermittelbar" gesetzt.
     *
     * @param stundenSatz Stundendatensatz für den ein Resultdata erzeugt wird.
     *
     * @return Data.
     */
    private Data erzeugeVerkehrsDatenLangZeitIntervallGeschwindigkeitVerteilungDatensatzMitVorbelegungImportdaten(Stundendatensatz stundenSatz) {
        Map<String, String> mapFahrzeugArtenWert                         = stundenSatz.getMapFZArtVerkehrsMenge();
        int                 vkLVo_size                                   = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("LVo").size();
        int                 vkSGV_size                                   = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("SGV").size();
        int                 vkBPA_size                                   = stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("BPA").size();
        Data                dataVerkehrsDatLangZeitIntGeschwVerteilungFS = _dav.createData(_atgVerkehrsDatLangZeitIntGeschwVerteilungFS);

        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getTextValue("T").setText("1 Stunde");

        // Nicht versorgte Werte:
        for (String wert : AR_VerkehrsDatenLangZeitIntervallGeschwVerteilungFSnichtImportierteAttribute) {
            dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getUnscaledValue(wert).setText(NICHT_ERMITTELBAR);
        }

        for (String fzArt : AR_vLVoSGvBPa) {
            String value = mapFahrzeugArtenWert.get(fzArt);

            if (value.equals(NICHT_ERMITTELBAR)) {
                dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getTextValue(fzArt).setText(value);
            } else {
                String doubleValueAsString = value.replace(',', '.');
                double doubleValue         = Double.parseDouble(doubleValueAsString);

                dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getUnscaledValue(fzArt).set(doubleValue);
            }
        }

        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray(Q_LVoGeschwKlasse).setLength(vkLVo_size);

        for (int i = 0; i < vkLVo_size; i++) {
            dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getUnscaledArray(Q_LVoGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("LVo").get(AR_lVoValues[i]));
        }

        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray(Q_SGVGeschwKlasse).setLength(vkSGV_size);

        for (int i = 0; i < vkSGV_size; i++) {
            dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getUnscaledArray(Q_SGVGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("SGV").get(AR_sGVValues[i]));
        }

        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray(Q_BPAGeschwKlasse).setLength(vkBPA_size);

        for (int i = 0; i < vkBPA_size; i++) {
            dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getUnscaledArray(Q_BPAGeschwKlasse).getValue(i).setText(stundenSatz.getGKlasseVerkehrsMengeGKlasse().get("BPA").get(AR_bPAValues[i]));
        }

        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray("vGrenzLVo").setLength(0);
        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray("vGrenzSGV").setLength(0);
        dataVerkehrsDatLangZeitIntGeschwVerteilungFS.getArray("vGrenzBPA").setLength(0);

        return dataVerkehrsDatLangZeitIntGeschwVerteilungFS;
    }

    /**
     * Generiere den SendeErlaubnis-Schlüssel.
     *
     * @param so SystemObject.
     * @param dd DataDescription.
     *
     * @return SendeErlaubnis-Schlüssel.
     */
    private String generiereSendeErlaubnisSchluessel(SystemObject so, DataDescription dd) {
        return String.format("%s#%s#%s", so.getPidOrId(), dd.getAttributeGroup().getPidOrId(), dd.getAspect().getPidOrId());
    }

    //~ GET METHODEN ==========================================================

    /**
     * Get-Methode für die TESTZWECKE.
     *
     * @return LangZeitZaehlStelle.
     */
    public LangZeitZaehlStelle getLzz() {
        return _lzz;
    }

    //~ METHODEN ==============================================================

    @Override
    public void initialize(ClientDavInterface dav) throws Exception {

        // Initialisiere ClientDavInterface:
        _dav = dav;
        NetzClientDavConnection.initDav(dav);

        // Initialisiere BAStStatusDaten:
        _bastStatusDaten = new SendeBAStStatusDaten(_typBAStBand, false);
    }

    //~ GET METHODEN ==========================================================

    @Override
    public boolean isRequestSupported(SystemObject systemObject, DataDescription dataDescription) {
        return true;
    }

    //~ METHODEN ==============================================================

    /**
     * Die Methode liest die Date mit dem übergebenen Dateinamen vom FTP-Server ein.
     *
     * Methode ist public für die TESTZWECKE.
     *
     * @param dateiName der Datei die importiert werden soll.
     * @param dateiStream Inhalt der Datei.
     *
     * @return instanz BastBestandsbandformat mit allen relevanten Attributen.
     */
    public BastBestandsbandformat konvertiereDateiInBastFormat(String dateiName, InputStream dateiStream) {
        ImportLineSplitterBASt lineSplitter       = new ImportLineSplitterBASt(new BastBestandsbandformat(), _bastStatusDaten);
        BastBestandsbandformat bastBestandsFormat = new BastBestandsbandformat();

        try {
            BufferedReader br    = new BufferedReader(new InputStreamReader(dateiStream, "US-ASCII"));
            int            count = 1;

            try {
                String line;

                while ((line = br.readLine()) != null) {
                    if (count == 1) {
                        if (_bastFormat == BAST_FORMAT.V2004) {
                            _lzz = lineSplitter.parseErsteHeaderLineV2004(line, dateiName);
                        } else {
                            _lzz = lineSplitter.parseErsteHeaderLineV2007(line, dateiName);
                        }

                        if (_lzz == null) {
                            _abbruchImport = true;
                        }

                        count++;
                    } else if ((count == 2)
                               &&!_abbruchImport) {
                        if (_bastFormat == BAST_FORMAT.V2004) {
                            _abbruchImport = (lineSplitter.parseZweiteHeaderLine2004(line, dateiName));
                        } else {
                            _abbruchImport = (lineSplitter.parseZweiteHeaderLine2007(line, dateiName));
                        }

                        count++;
                    } else if ((count == 3)
                               &&!_abbruchImport) {
                        if (_bastFormat == BAST_FORMAT.V2004) {
                            _abbruchImport = (lineSplitter.parseDritteHeaderLine2004(line, dateiName));
                        } else {
                            lineSplitter.parseDritteHeaderLine2007(line, dateiName);
                        }

                        count++;
                    } else if (!_abbruchImport) {
                        if (_bastFormat == BAST_FORMAT.V2004) {
                            bastBestandsFormat = lineSplitter.parseStundenDatensaetze2004(line);
                        } else {
                            bastBestandsFormat = lineSplitter.parseStundenDatensaetzeKlassifizierungV2007(line);
                        }
                    }
                }
            }
            finally {
                br.close();
            }
        }
        catch (IOException e) {
            sendeMeldungFehler("100", String.format("Es sind Probleme beim einlesen der Daten aufgetreten: %s", dateiName));
            debug.error(String.format("Es sind Probleme beim einlesen der Daten aufgetreten: %s <%s>", dateiName, e.getMessage()), e.getStackTrace());
        }

        return bastBestandsFormat;
    }

    /**
     * Methode die das Data Objekt in Dav-Daten transformiert.
     *
     * @param data Dataobjekt.
     */
    private void konvertiereDavDatenImport(Data data) {

        //
        if (data.getTextValue("FTP-Server").getText().equals("IMPORT-ABBRUCH")) {
            _abbruchImport = true;
        } else {
            _abbruchImport = false;

            Data.TextValue bastBandAusgabeFormat = data.getTextValue("BastBandAusgabeFormat");

            _bastFormat = BAST_FORMAT.ermittleFormat(bastBandAusgabeFormat.getText());

            // DaV-Datenformat: Siehe Methode de.inovat.buv.plugin.gtm.bast.gui.ImportGUIVerwaltung.sendeImportDaten()
            //
            // "-Benutzer:%s-Passwort:%s-URL:%s-Port:%s-Passivmodus:%s-Servertyp:%s"
            //
            // alte Variante (ohne Port und FTP/SFTP):
            // "Benutzer:%sPasswort:%sURL:%sPassivmodus:%s"
            //
            // FTP-Server Angaben:
            String ftpServer = data.getTextValue("FTP-Server").getText();

            _ftpBenutzer   = "";
            _ftpPasswd     = "";
            _ftpServerName = "";
            _ftpServerPort = 0;

            String passivmodus = "";
            String servertyp   = "";

            //
            Pattern pattern = Pattern.compile("-Benutzer:(.*)-Passwort:(.*)-URL:(.*)-Port:(.*)-Passivmodus:(.*)-Servertyp:(.*)");
            Matcher matcher = pattern.matcher(ftpServer);

            if (matcher.find()) {
                _ftpBenutzer   = matcher.group(1);
                _ftpPasswd     = matcher.group(2);
                _ftpServerName = matcher.group(3);

                try {
                    _ftpServerPort = Integer.parseInt(matcher.group(4));
                }
                catch (Exception e) {
                    debug.error("Es sind Probleme bem Auslesen des FTP-Server-Port aus dem DaV-Attribut aufgetreten.");
                }

                passivmodus = matcher.group(5);
                servertyp   = matcher.group(6);
            }

            // alte Variante:
            else {
                pattern = Pattern.compile("Benutzer:(.*)Passwort:(.*)URL:(.*)Passivmodus:(.*)");
                matcher = pattern.matcher(ftpServer);

                if (matcher.find()) {
                    _ftpBenutzer   = matcher.group(1);
                    _ftpPasswd     = matcher.group(2);
                    _ftpServerName = matcher.group(3);
                    _ftpServerPort = 21;
                    passivmodus    = matcher.group(4);
                    servertyp      = "FTP";
                } else {
                    debug.error("Es sind Probleme bem Auslesen der FTP-Server-Einstellungen aus dem DaV-Attribut aufgetreten.");
                }
            }

            _ftpPassivModus = (passivmodus.toLowerCase().equals("ja"));
            _istFtp         = servertyp.toUpperCase().equals("FTP");

            // LangZeitZählStellen.
            Data.TextArray   txtAr_Dateien      = data.getItem("ImportDateien").asTextArray();
            Data.TextValue[] txtValueAr_Dateien = txtAr_Dateien.getTextValues();

            // Füllen der Liste mit allen PID-Nummern der zu importierenden Langzeitzählstellen.
            _importDateienList = new ArrayList<String>();

            for (Data.TextValue dateiName : txtValueAr_Dateien) {
                _importDateienList.add(dateiName.getText());
            }
        }
    }

    /**
     * Methode die das Data-Objekt transformiert. Das data-Objekt enthält einen String, in dem alle Informationen durch
     * ein @ angehängt sind. Dabei sind folgende Informationen enthalten: <li> dateiname </li> <li> url </li> <li> user
     * </li> <li> passwd </li> <li> bastFormat </li>
     *
     * @param data Dataobjekt.
     */
    private void konvertiereDavDatenVerzeichnisueberwachungs(Data data) {

        // DaV-Datenformat: Siehe Methode de.inovat.kex.ftp.DavKommunikation.sendeDavDaten()
        // ("%s@%s@%s@%s@%s@%s@%s@%s@", dateiName, url, port, user, passwd, passivModus, bastFormat, serverTyp);
        //
        String   allInfos    = data.getTextValue("ImportDatei").getText();
        String[] einzelInfos = allInfos.split("@");

        //
        _importDateienList = new ArrayList<String>();
        _importDateienList.add(einzelInfos[0]);

        //
        _bastFormat = BAST_FORMAT.ermittleFormat(Integer.parseInt(einzelInfos[6]));

        //
        _ftpServerName  = einzelInfos[1];
        _ftpServerPort  = Integer.parseInt(einzelInfos[2]);
        _ftpBenutzer    = einzelInfos[3];
        _ftpPasswd      = einzelInfos[4];
        _ftpPassivModus = (einzelInfos[5].toLowerCase().equals("ja"));

        // FTP/SFTP:
        _istFtp = !einzelInfos[7].toUpperCase().equals("SFTP");
    }

    // Main-Methode
    public static void main(String[] args) {
        try {
            BAStImportSteuerung bastImport = new BAStImportSteuerung();

            StandardApplicationRunner.run(bastImport, args);

            // Debug erst hier initialisieren, da erst ab hier verfügbar !!
            debug = Debug.getLogger();

            //
            bastImport.anmeldenDavEmpfaengerBAStImport();
            bastImport.sendeMeldungInfo("100", "Der Prozess BASt-Band läuft. Der Import kann gestartet werden.");

            /*
            // ---------------------------------------------------------------
            // Test-Source für das Auslesen einer Bast-Datei
            // ---------------------------------------------------------------
            // Parameter (Bast-Format und Datei):
            bastImport._bastFormat = BAST_FORMAT.V2004;

            String      dateiName   = "C:\\BW-DAUZ\\BW72201007.16c";  // hier die zu importierte Bast-Datei angeben
            InputStream dateiStream = new FileInputStream(dateiName);

            // Bast-Datei lesen:
            BastBestandsbandformat bastBestandsFormat = bastImport.konvertiereDateiInBastFormat(dateiName, dateiStream);

            System.out.println("Test: Datei gelesen");

            // Ausgelesene Bast-Daten bewrten:
            // Fahrstreifen fs;
            // bastImport.erzeugeDatensatzV2004(bastBestandsFormat, fs, 0);
            // System.out.println("Test: Ende");
            // ---------------------------------------------------------------
            */
        }
        catch (Exception e) {
            debug.error("Es ist ein Fehler aufgetreten. Applikation wird beendet.", e);
            System.exit(2);
        }
    }

    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        String par;

        // --------------------------------------------------------------------
        // Hole den Parameter für das SystemObjekt der BAStImportSteuerung:
        par = "-pidBastExportImportSteuerung";

        try {
            _typBAStBand = argumentList.fetchArgument(par).asNonEmptyString();
        }
        catch (Exception e) {
            debug.error(String.format("Bitte den Parameter <%s> für das SystemObjekt der BAStImportSteuerung übergeben!", par), e);

            throw new RuntimeException(e);
        }

        // --------------------------------------------------------------------
        // Hole den Parameter für den lokalen Temp-Ordner:
        par = "-tempVerzeichnis";

        try {
            _tempOrdner = argumentList.fetchArgument(par).asNonEmptyString();
        }
        catch (Exception e) {

            // DefaultWert:
            _tempOrdner = "../../TempOrdnerFtp/";

            //
            debug.warning(String.format(                                                                                             //
                "Der Parameter <%s> für den lokalen Temp-Ordner wurde nicht übergeben! ===> wird mit dem Wert <%s> initialisiert.",  //
                par, _tempOrdner));
        }

        // Erstelle den angegebenen Temp-Ordner, falls noch nicht existiert:
        File file = new File(_tempOrdner);

        if (!file.exists()) {
            try {
                file.mkdirs();
            }
            catch (SecurityException e) {
                debug.error(String.format("Der Ordner <%s> kann nicht erstellt werden!", _tempOrdner), e);

                throw new RuntimeException(e);
            }
        }
    }

    /**
     * In der run Methode des Threads BAStBandImport wird der Import der Datei durchgeführt. Folgende Schritte sind
     * dabei zu vollziehen, nachdem eine erfolgreiche Verbindung zum FTP-Server statt gefunden hat. Sollte dies nicht
     * möglich sein, wird eine Statusmeldung versendet. Für alle Dateien die importiert werden sollen:
     * <ul>
     * <li>0. Sperren der Datei via DATEI_LOCK Endung </li>
     * <li>1. Lese Datei vom FTP-Server ein </li>
     * <li>2. Konvertiere die Zeilen in das interne DaV-Format. </li>
     * <li>3. Schreibe Datensatz in DaV.</li>
     * <li>4. Bestätige Import der Datei via Statusmeldung. </li> <li>5. Benenne Datei in bearbeitet um. </li>
     * </ul>
     */
    public void run() {

        // -------------------------------------------------------------
        // Initialisiere die Server-Kommunikationsklasse:
        // FTP
        if (_istFtp) {
            _serverKommunikation = new FTPServerKommunikation();
        }

        // SFTP:
        else {
            _serverKommunikation = new SFTPServerKommunikation();
        }

        _serverKommunikation.parameterUebergabe(_tempOrdner, _ftpServerName, _ftpServerPort, _ftpBenutzer, _ftpPasswd, _ftpPassivModus, _bastFormat._ftpImportOrdner);
        _prozentImport = 1;
        sendeMeldungInfo("1", "Der Importvorgang wird gestartet...");

        // -------------------------------------------------------------
        // Erstelle die Server-Verbindung:
        Exception exception = _serverKommunikation.verbindeMitServer();

        // Fehlermeldung:
        if (exception != null) {
            debug.error(exception.getMessage(), exception.getStackTrace());
            abbrechen(exception.getMessage());
        }

        // OK:
        else {

            // Meldungen:
            sendeMeldungInfo("", "Die Verbindung zum FTP-Server wurde hergestellt bzw. aktualisiert.");
            sendeMeldungInfo("", "Der Login auf dem FTP-Server ist erfolgt.");
        }

        if (exception == null) {

            // -------------------------------------------------------------
            // Bearbeite alle zu importierte Dateien:
            for (String dateiName : _importDateienList) {
                if (!_abbruchImport) {

                    // -------------------------------------------------------------
                    // Definiere Dateipfad:
                    String dateiMitPfad;

                    if (_bastFormat._ftpImportOrdner.isEmpty()) {
                        dateiMitPfad = dateiName;
                    } else {
                        dateiMitPfad = String.format("%s%s%s", _bastFormat._ftpImportOrdner, FILESEPARATOR_SERVER, dateiName);
                    }

                    // -------------------------------------------------------------
                    // Sperre die Datei:
                    exception = _serverKommunikation.sperreDatei(dateiMitPfad);

                    if (exception == null) {

                        // -------------------------------------------------------------
                        // Lese die Datei:
                        InputStream dateiStream = _serverKommunikation.ermittleDateiInputStream(String.format("%s%s", dateiMitPfad, DATEI_LOCK));

                        if (dateiStream != null) {

                            // -------------------------------------------------------------
                            // Konvertiere die Datei in BAST-Format:
                            BastBestandsbandformat bastBestandsFormat = konvertiereDateiInBastFormat(dateiName, dateiStream);

                            if (!_abbruchImport) {

                                // -------------------------------------------------------------
                                // Sende die Datei an DaV:
                                boolean istGesendet;

                                try {
                                    sendeImportiertenDatenFuerAlleFS(bastBestandsFormat);
                                    istGesendet = true;
                                }
                                catch (Exception e) {
                                    istGesendet = false;
                                    debug.error(String.format("Es sind Probleme beim senden der Daten aufgetreten: %s <%s>", dateiName, e.toString()),  //
                                                e.getStackTrace());
                                }

                                if (!istGesendet) {
                                    sendeMeldungFehler("100", String.format("Es sind Probleme beim senden der Daten aufgetreten: %s", dateiName));
                                } else if (!_abbruchSystem) {

                                    // -------------------------------------------------------------
                                    // Sperre der Datei aufheben:
                                    exception = _serverKommunikation.sperreDateiAufheben(dateiMitPfad);

                                    if (exception == null) {
                                        sendeMeldungInfo("", String.format("Die Datei %s wurde erfolgreich importiert.", dateiName));
                                    } else {
                                        sendeMeldungFehler("", String.format(                                                                           //
                                            "Die Datei %s wurde importiert, die Sperre der Datei konnte aber nicht aufgehoben werden,", dateiName));
                                    }
                                }
                            }
                        } else {
                            debug.error(String.format("Es sind Probleme beim einlesen der Daten aufgetreten: %s", dateiName));
                            sendeMeldungFehler("100", String.format(                                                                                    //
                                "Es sind Probleme beim einlesen der Daten <%s> aufgetreten, eventuell ist dies keine Datei sondern ein Verzeichnis!?",  //
                                dateiName));
                        }
                    } else {
                        sendeMeldungFehler("100", exception.getMessage());
                        debug.error(String.format("Die Datei <%s> konnte nicht gesperrt werden.%n<%s>",                                                 //
                                                  dateiName, exception.getMessage()), exception.getStackTrace());
                    }
                }

                // -------------------------------------------------------------
                // Abbruch:
                else {
                    sendeMeldungInfo("", " Der Import wurde abgebrochen.");
                    stopThread("Der Importvorgang wurde abgeborchen.");

                    break;
                }
            }

            // -------------------------------------------------------------
            // Finalmeldungen:
            if ((_prozentImport < 100)
                &&!_abbruchImport) {
                _prozentImport = 100;
                sendeMeldungInfo((new Integer(_prozentImport)).toString(), " Der Importvorgang  wurde abgeschlossen.");
            }

            stopThread("Es kann ein neuer Import gestartet werden.");
        }

        // -------------------------------------------------------------
        // Trenne die Verbindung:
        _serverKommunikation.trenneVerbindungMitServer();
    }

    /**
     * Das Senden der DaV-Daten.
     *
     * @param listeResultData Liste der ResultData die versendet werden.
     */
    private void sendeDavDaten(ArrayList<ResultData> listeResultData) {

        // Nur wenn der Status ok ist, sollen die Daten importiert werden.
        // Wenn dies nicht der Fall ist, wird die globale Variable _abbruchSystem auf true gesetzt, um
        // mögliche weitere bestimmungen der Daten für die Importdaten zu verhindern.
        if (listeResultData.size() > 0) {
            ResultData[] results = new ResultData[listeResultData.size()];
            int          i       = 0;

            for (ResultData data : listeResultData) {
                results[i] = data;
                i++;
            }

            try {

                // Warte auf SendenErlaubnis:
                Boolean sendeErlaubnis = _mapSendeErlaubnis.get(generiereSendeErlaubnisSchluessel(results[0].getObject(), results[0].getDataDescription()));

                if ((sendeErlaubnis == null)
                    || (!sendeErlaubnis)) {
                    try {
                        Thread.sleep(2000);
                    }
                    catch (InterruptedException e) {
                        debug.error("");
                    }
                }

                // Sende die Daten:
                _dav.sendData(results);
            }
            catch (Exception e) {
                debug.error("Die Daten konnten nicht versendet werden.", e);
            }
        }
    }

    /**
     * Die Methode sendet die importierten Verkehrsdaten.
     *
     * @param bastBestandsFormat BastBestandsBandformat.
     */
    private void sendeImportiertenDatenFuerAlleFS(BastBestandsbandformat bastBestandsFormat) {

        // -------------------------------------------------------------------------------
        // Ermittle alle FS:
        List<Fahrstreifen> listeFs = new ArrayList<Fahrstreifen>();

        listeFs.addAll(Arrays.asList(_lzz.getInAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen()));
        listeFs.addAll(Arrays.asList(_lzz.getGegenAuswertequerschnitt().getMessQuerschnitt().getFahrstreifen()));

        for (Fahrstreifen fs : listeFs) {
            if (!_abbruchSystem) {
                int prozentSchritt = bestimmeProzentSchritt();

                // -------------------------------------------------------------------------------
                SystemObject soFahrstreifen = _dav.getDataModel().getObject(fs.getConfigurationObject().getPid());

                // -------------------------------------------------------------------------------
                // V2004
                // -------------------------------------------------------------------------------
                if (_bastFormat == BAST_FORMAT.V2004) {

                    // -------------------------------------------------------------------------------
                    // Anmeldung beim Datenverteiler als Sender für die Attributgruppe bzgl. BASt-Band Import V2004
                    try {
                        _dav.subscribeSender(this, soFahrstreifen, _datenbeschreibungVerkehrsDatenLangZeitIntervall, SenderRole.source());
                    }
                    catch (OneSubscriptionPerSendData ignored) {}

                    // -------------------------------------------------------------------------------
                    try {
                        erzeugeDatensatzV2004(bastBestandsFormat, fs, prozentSchritt);
                    }
                    catch (Exception e) {
                        debug.error("Es sind Probleme beim Import aufgetreten.", e);
                    }

                    // -------------------------------------------------------------------------------
                    // Abmelden vom Datenverteiler des übergebenem Systemobjektes
                    try {
                        _dav.unsubscribeSender(this, soFahrstreifen, _datenbeschreibungVerkehrsDatenLangZeitIntervall);
                    }
                    catch (Exception e) {
                        debug.warning("Die Abmeldung für den Fahrstreifen bzgl. Senden von VerkehrsDatenLangZeitIntervall konnte nicht vollzogen werden. ", e);
                    }
                }

                // -------------------------------------------------------------------------------
                // V2007
                // -------------------------------------------------------------------------------
                else {

                    // -------------------------------------------------------------------------------
                    // Anmelden für das Versenden der Daten der Attributgruppe TLSLveErgebnismeldungVersion24 V2007
                    try {
                        _dav.subscribeSender(this, soFahrstreifen, _datenbeschreibungVerkehrsDatLangZeitIntGeschwVerteilungFS, SenderRole.source());
                    }
                    catch (OneSubscriptionPerSendData ignored) {}

                    // -------------------------------------------------------------------------------
                    try {
                        erzeugeDatensatzV2007(bastBestandsFormat, fs, prozentSchritt);
                    }
                    catch (Exception e) {
                        debug.error("Es sind Probleme beim Import aufgetreten.", e);
                    }

                    // -------------------------------------------------------------------------------
                    // Abmelden vom Datenverteiler des übergebenem Systemobjektes
                    try {
                        _dav.unsubscribeSender(this, soFahrstreifen, _datenbeschreibungVerkehrsDatLangZeitIntGeschwVerteilungFS);
                    }
                    catch (Exception e) {
                        debug.warning("Die Abmeldung konnte nicht vollzogen werden. ", e);
                    }
                }
            } else {
                break;
            }
        }
    }

    /**
     * Die Methode sendet die übergebene Fehler_Meldung an den DaV.
     *
     * @param prozent Prozent.
     * @param meldung Fehlermeldung die publiziert wird.
     */
    private void sendeMeldungFehler(String prozent, String meldung) {
        try {
            _bastStatusDaten.sendeBAStStatusDaten(prozent, meldung);
            debug.error(meldung);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Es sind Probleme beim Senden der Statusinformationsdaten aufgetreten.", e);
        }
    }

    /**
     * Die Methode sendet die übergebene Info_Meldung an den DaV.
     *
     * @param prozent Prozent.
     * @param meldung Infomeldung die publiziert wird.
     */
    private void sendeMeldungInfo(String prozent, String meldung) {
        try {
            _bastStatusDaten.sendeBAStStatusDaten(prozent, meldung);
            debug.info(meldung);
        }
        catch (SendSubscriptionNotConfirmed e) {
            debug.error("Es sind Probleme beim Senden der Statusinformationsdaten aufgetreten.", e);
        }
    }

    /**
     * Methode für die TESTZWECKE.
     *
     *     @param bastFormat BASt-Format.
     *     @param dav DaV.
     */
    public void setzeDatenFuerTestzwecke(BAST_FORMAT bastFormat, ClientDavInterface dav) {
        _istFuerTestzwecke                 = true;
        _bastFormat                        = bastFormat;
        _dav                               = dav;
        _atgVerkehrsDatenLangZeitIntervall = _dav.getDataModel().getAttributeGroup(ATG_VERKEHRS_DATEN_LANG_ZEITINTERVALL);
    }

    /**
     * Die Methode erzeugt einen Thread, der den BastBandImport startet. Durch die Verwendung eines Threads ist es
     * möglich den Import zu jedem Zeitpunkt zu unterbrechen.
     */
    private void starteThread() {
        if (!_threadImportAktiv) {
            _threadBastBandImport = new Thread(this);
            _threadBastBandImport.setName("BAStBandImport");
            _threadImportAktiv = true;
            _threadBastBandImport.start();
        }
    }

    /**
     * Stoppt den Thread.
     *
     * @param meldung Statusmeldung die an den Clienten via DaV geschickt wird.
     */
    private void stopThread(String meldung) {
        if (_threadImportAktiv) {
            _threadImportAktiv = false;
            sendeMeldungInfo("100", meldung);
        }
    }

    /**
     * Update-Methode wird vom Datenverteiler aufgerufen. Die übergebenen Datensätze werden verarbeitet.
     *
     * @param resultDatas Resultdata, entspricht den Attributen eines BASt-Band importes, d.h. Referenz auf
     *                    Langzeitzählstellen username passwort url zum FTP-Server auf dem die Importierten Daten
     *                    abgelegt werden.
     */
    @Override
    public void update(ResultData[] resultDatas) {
        _abbruchImport = false;

        for (ResultData result : resultDatas) {
            if (result.hasData()) {
                try {

                    // ----------------------------------------------------------------------
                    // solange ein import Thread läuft, soll kein neuer gestartet werden. Da es sonst
                    // zu Problemen bei der Datenverarbeitung kommen kann.
                    while (_threadImportAktiv) {
                        try {
                            Thread.sleep(100);
                        }
                        catch (InterruptedException e) {
                            debug.info("Es hat Probleme beim Pausieren des Threads gegeben.", e);
                        }
                    }

                    // ----------------------------------------------------------------------
                    // Entscheide für die unterschiedlich angemeldeten Datenbeschreibungen
                    // wie sie intern konvertiert werden müssen:
                    Data data = result.getData();

                    if (result.getDataDescription().getAttributeGroup().getPidOrNameOrId().equals(ATG_VERZEICHNIS_UEBERWACHUNG)) {
                        konvertiereDavDatenVerzeichnisueberwachungs(data);
                    } else {
                        konvertiereDavDatenImport(data);
                    }

                    // ----------------------------------------------------------------------
                    // Falls nicht Abbruch:
                    if (!_abbruchImport) {

                        // ----------------------------------------------------------------------
                        // Ungültiges Format --> Abbruch:
                        if ((_bastFormat == null)
                            || (_bastFormat == BAST_FORMAT.AXL_EF1)) {
                            sendeMeldungFehler("", "Das übergebene BASt-Band Format existiert nicht, der Import wird abgebrochen.");

                            return;
                        }

                        // ----------------------------------------------------------------------
                        try {
                            starteThread();
                        }
                        catch (Exception e) {
                            String infoText = "Es sind Probleme beim Starten des Threads aufgetreten. Die Anwendung wird geschlossen. ";

                            debug.error(infoText, e);
                            abbrechen(infoText);
                        }
                    }
                }
                catch (Exception e) {
                    debug.error("Die Daten konnten nicht in internes Format transformiert werden.", e);
                }
            }
        }
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 23.02.17
