/*
 * Copyright 2008 by inovat, Dipl.-Ing. H. C. Kniß
 * ALL RIGHTS RESERVED.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL inovat OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



package de.inovat.kex.bast.importer;

//~ JDK IMPORTE ===============================================================

import java.io.InputStream;

//~ INTERFACES ================================================================

/**
 * Interface stellt die Schnittstelle für die Server-Kommunikation beim BASt-Band-Import
 * dar (s. {@link de.inovat.kex.bast.importer.BAStImportSteuerung#run()}).
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Liliya Givorgizova (LG)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public interface IServerKommunikation {

    /**
     * Ermittle den Inhalt der angegeben Server-Datei.
     *
     * @param dateiPfad Server-Datei mit dem Pfad.
     *
     * @return Inhalt der Server-Datei als InputStream oder null.
     */
    public InputStream ermittleDateiInputStream(String dateiPfad);

    /**
     * Methode dient für die Übergabe der Parameter.
     *
     * @param tempOrdner Lokaler Temp-Ordner.
     * @param host Server-Name.
     * @param port Server-Port.
     * @param benutzer Server-Benutzer.
     * @param passwort Passwort.
     * @param passivModus Legt fest, ob die Server-Verbindung im Passiv-Modus eingebaut werden soll.
     * @param ordner Server-Ordner, wo die zu bearbeitenden Dateien liegen. .
     */
    public void parameterUebergabe(String tempOrdner, String host, int port, String benutzer, String passwort, boolean passivModus, String ordner);

    /**
     * Die Methode sperrt die Datei, die im folgenden Schritt eingelesn wird.
     * Folgende Ausnahmen werden dabei behandelt:
     * <ul>
     * <li>
     *     Die Datei ist schon gesperrt (kenntlich durch den suffix {@link BAStImportSteuerung#DATEI_LOCK}),
     *     der Benutzer erhält darüber eineStatusmeldung. </li>
     * <li>
     *     Die Datei ist schon bearbeitet (kenntlich durch den suffix {@link BAStImportSteuerung#DATEI_BEARBEITET}),
     *     der Benutzer erhält darüber eine Statusmeldung. </li>
     * <li>
     *     Die Datei existiert nicht. Der Benutzer erhält darüber eine Statusmeldung. </li>
     *     </ul>
     *
     * @param dateiMitPfad Dateiname und Pfad der Datei, die gesperrt wird.Bei der Sperrung handelt
     *                     es sich um ein Hinzufügen des Suffix {@link BAStImportSteuerung#DATEI_LOCK}.
     *
     * @return Exception, falls einen _Fehler aufgetretten ist oder
     * null wenn die Umbenennung geklappt hat.
     */
    public Exception sperreDatei(String dateiMitPfad);

    /**
     * Die Methode hebt die Sperre der Datei wieder auf und benennt die Datei um
     * mit Suffix {@link BAStImportSteuerung#DATEI_BEARBEITET} wobei vorher
     * der Suffix {@link BAStImportSteuerung#DATEI_LOCK} eliminiert wird.
     *
     * @param dateiMitPfad Dateiname und Pfad der Datei, die gesperrt wurde und deren Sperre aufgehoben wird.
     *
     * @return Exception, falls einen _Fehler aufgetretten ist oder
     * oder null, wenn die Sperraufhebung geklappt hat, false sonst.
     */
    public Exception sperreDateiAufheben(String dateiMitPfad);

    /**
     * Trenne die Verbindung mit dem Server.
     *
     * @return Exception, falls einen _Fehler aufgetretten ist
     *  oder null - wenn die Verbindung zum Server erfolgreich getrennt wird.
     */
    public Exception trenneVerbindungMitServer();

    /**
     * Verbinde mit dem Server.
     *
     * @return Exception, falls einen _Fehler aufgetretten ist
     *  oder null - wenn die Verbindung zum Server erfolgreich aufgebaut wird.
     */
    public Exception verbindeMitServer();
}


//~Formatiert mit 'inovat Kodierkonvention' am 27.06.13
